from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict

from layers_linter.config import LayerConfig, LibConfig, load_config
from layers_linter.imports import collect_imports
from layers_linter.search_modules import (
    ModulePathT,
    find_modules_in_directory,
    FilePathT,
    match_pattern,
)


@dataclass
class Problem:
    line_number: int
    module_path: str
    imported_module: str
    code: str

    @property
    def message(self) -> str:
        raise NotImplementedError("Subclasses must implement this property")

    @property
    def file_path(self) -> str:
        return self.module_path.replace(".", "/") + ".py"


@dataclass
class LayerProblem(Problem):
    layer_from: str
    layer_to: str

    @property
    def message(self) -> str:
        return (
            f"Invalid layer dependency: '{self.layer_from or 'not-defined'}' "
            f"-> '{self.layer_to or 'not-defined'}'"
        )

    def __str__(self):
        return f"{self.file_path}:{self.line_number}: {self.message}"


@dataclass
class LibProblem(Problem):
    lib_name: str
    layers: List[str]

    @property
    def message(self) -> str:
        layers_str = ", ".join(self.layers)
        return f"Layers [{layers_str or 'not-defined'}] cannot use restricted library '{self.lib_name}'"

    def __str__(self):
        return f"{self.file_path}:{self.line_number}: {self.message}"


@dataclass
class NoLayerProblem(Problem):
    @property
    def message(self) -> str:
        return f"Module '{self.module_path}' does not belong to any layer"

    def __str__(self):
        return f"{self.file_path}: {self.message}"


def analyze_dependencies(
    project_root: Path,
    layers: Dict[str, LayerConfig],
    libs: Dict[str, LibConfig],
    exclude_modules: List[str],
    check_no_layer: bool = False,
) -> List[Problem]:
    modules_list = find_modules_in_directory(
        FilePathT(project_root), patterns=None, exclude_patterns=exclude_modules
    )
    all_project_modules = set(module_path for _, module_path in modules_list)

    module_to_layers: Dict[ModulePathT, List[str]] = defaultdict(list)
    for _, module_path in modules_list:
        for layer_name, layer_info in layers.items():
            # Check if module should be excluded for this layer
            excluded = False
            if layer_info.exclude_modules:
                for pattern in layer_info.exclude_modules:
                    if match_pattern(module_path, pattern):
                        excluded = True
                        break

            if excluded:
                continue

            for pattern in layer_info.contains_modules:
                if match_pattern(module_path, pattern):
                    module_to_layers[module_path].append(layer_name)
                    break

    # Check for modules that belong to multiple layers
    for module_path, module_layers in module_to_layers.items():
        if len(module_layers) > 1:
            layers_str = ", ".join(module_layers)
            raise ValueError(f"Module '{module_path}' belongs to multiple layers: [{layers_str}]")

    module_imports = collect_imports(all_project_modules, modules_list)

    problems = []

    for module_path, imports in module_imports.items():
        layers_a = module_to_layers.get(module_path, [])
        for import_info in imports:
            imported_module = import_info.module
            lineno = import_info.line_number

            if not import_info.is_internal:
                continue

            layers_b = module_to_layers.get(imported_module, [])

            if not layers_a or not layers_b:
                continue

            allowed = False
            for la_name in layers_a:
                for lb_name in layers_b:
                    la = layers[la_name]

                    # Depends_on check: layer A can only depend on those specified in depends_on.
                    depends_on_ok = True
                    if la.depends_on is not None:
                        if lb_name not in la.depends_on:
                            depends_on_ok = False

                    if depends_on_ok:
                        allowed = True
                        break
                if allowed:
                    break

            if not allowed:
                for la_name in layers_a:
                    for lb_name in layers_b:
                        problems.append(
                            LayerProblem(
                                line_number=lineno,
                                module_path=module_path,
                                imported_module=imported_module,
                                layer_from=la_name,
                                layer_to=lb_name,
                                code="LA001",
                            )
                        )

    # Check library dependencies
    for module_path, imports in module_imports.items():
        layers_a = module_to_layers.get(module_path, [])

        for import_info in imports:
            imported_module = import_info.module
            lineno = import_info.line_number

            # Skip if it's not a library import
            if import_info.is_internal:
                continue

            # Find if this is a known library
            matching_libs = []
            for lib_name in libs:
                if imported_module == lib_name or imported_module.startswith(f"{lib_name}."):
                    matching_libs.append(lib_name)

            for lib_name in matching_libs:
                lib_config = libs[lib_name]

                # If library has no restrictions, skip
                if lib_config.allowed_in is None:
                    continue

                # Check if any of the module's layers are allowed to use this library
                allowed = False
                for layer_name in layers_a:
                    if layer_name in lib_config.allowed_in:
                        allowed = True
                        break

                if not allowed:
                    problems.append(
                        LibProblem(
                            line_number=lineno,
                            module_path=module_path,
                            imported_module=imported_module,
                            code="LA020",
                            lib_name=lib_name,
                            layers=layers_a,
                        )
                    )

    if check_no_layer:
        for file_path, module_path in modules_list:
            if module_path not in module_to_layers:
                problems.append(
                    NoLayerProblem(
                        line_number=0,
                        module_path=module_path,
                        imported_module="",
                        code="LA002",
                    )
                )

    return problems


def run_linter(project_root: Path, config_path: Path, check_no_layer: bool = False) -> List[Problem]:
    layers, libs, exclude_modules = load_config(config_path)
    return analyze_dependencies(project_root, layers, libs, exclude_modules, check_no_layer)



import ast
from collections import defaultdict
from dataclasses import dataclass
from typing import Set, List, Dict

from layers_linter.search_modules import ModulePathT


@dataclass
class ImportInfo:
    module: ModulePathT
    line_number: int
    is_internal: bool  # True if it's a project module, False if it's a library


class ImportVisitor(ast.NodeVisitor):
    def __init__(self, current_module: ModulePathT, all_project_modules: Set[ModulePathT]):
        self.imports: List[ImportInfo] = []
        self.current_module = current_module
        self.all_project_modules = all_project_modules
        self.inside_type_checking = False

    def process_import(self, node: ast.AST, module_name: ModulePathT):
        if self.inside_type_checking:
            return

        is_internal = module_name in self.all_project_modules
        self.imports.append(ImportInfo(module_name, node.lineno, is_internal))

    def visit_Import(self, node: ast.Import):
        for alias in node.names:
            name = alias.name
            self.process_import(node, ModulePathT(name))

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.level > 0:
            if not self.current_module:
                return
            parts = self.current_module.split(".")
            level = node.level
            if level > len(parts):
                return
            base_parts = parts[:-level]
            module = node.module
            if not module:
                return
            new_parts = base_parts + [module]
            module_name = ModulePathT(".".join(new_parts))
        else:
            module_name = ModulePathT(node.module)
        self.process_import(node, module_name)

    def visit_If(self, node: ast.If):
        if isinstance(node.test, ast.Name) and "TYPE_CHECKING" in node.test.id:
            old_flag = self.inside_type_checking
            self.inside_type_checking = True
            for stmt in node.body:
                self.visit(stmt)
            self.inside_type_checking = old_flag
            for stmt in node.orelse:
                self.visit(stmt)
        else:
            self.generic_visit(node)


def collect_imports(all_project_modules, modules_list) -> Dict[ModulePathT, List[ImportInfo]]:
    module_imports: Dict[ModulePathT, List[ImportInfo]] = defaultdict(list)
    for path, module_path in modules_list:
        with open(path) as f:
            content = f.read()
        tree = ast.parse(content)
        visitor = ImportVisitor(module_path, all_project_modules)
        visitor.visit(tree)
        module_imports[module_path].extend(visitor.imports)

    return module_imports



import warnings
from pathlib import Path

from layers_linter.analyzer import analyze_dependencies
from layers_linter.config import load_config


class LayersLinter:
    name = "layers-linter"
    version = "3.2.1"

    def __init__(self, tree, filename, lines, options):
        self.filename = Path(filename).resolve()
        self.options = options

    def run(self):
        config_path = getattr(self.options, "la_config", None)
        if not config_path:
            warnings.warn("Bounded contexts check skipped: config path not specified")
            return

        config_path = Path(config_path).resolve()
        if not config_path.exists() or not config_path.is_file():
            warnings.warn(
                f"Bounded contexts check skipped: config file {config_path} does not exist"
            )
            return

        if len(self.options.filenames) == 0:
            warnings.warn("Bounded contexts check skipped: no project root to check")
            return

        if len(self.options.filenames) > 1:
            warnings.warn("Bounded contexts check skipped: multiple files not supported")
            return

        layers, libs, exclude_modules = load_config(config_path)

        for problem in analyze_dependencies(
            self.options.filenames[0], layers, libs, exclude_modules, check_no_layer=True
        ):
            yield (problem.line_number, 0, f"{problem.code} {problem.message}", LayersLinter)

    @staticmethod
    def add_options(option_manager):
        option_manager.add_option(
            "--la-config",
            type=str,
            dest="la_config",
            help="Path to layers-linter configuration file (layers.toml)",
            parse_from_config=True,
            default="layers.toml",
        )



import tomllib
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional


@dataclass
class LayerConfig:
    contains_modules: List[str]
    depends_on: Optional[List[str]]
    exclude_modules: Optional[List[str]]


@dataclass
class LibConfig:
    allowed_in: Optional[List[str]]  # Layers allowed to use this lib


def load_config(
    config_path: Path,
) -> tuple[Dict[str, LayerConfig], Dict[str, LibConfig], List[str]]:
    with open(config_path, "rb") as f:
        raw_config = tomllib.load(f)

    layers_config = raw_config.get("layers", {})
    libs_config = raw_config.get("libs", {})
    exclude_modules = raw_config.get("exclude_modules", [])

    if not isinstance(exclude_modules, list):
        exclude_modules = []

    layers = {}
    for layer_name, layer_info in layers_config.items():
        depends_on = layer_info.get("depends_on", "none")
        if depends_on == "none":
            depends_on_parsed = None
        elif isinstance(depends_on, list):
            depends_on_parsed = depends_on
        else:
            depends_on_parsed = None

        contains_modules = layer_info.get("contains_modules", [])
        if not isinstance(contains_modules, list):
            contains_modules = []

        exclude_modules_layer = layer_info.get("exclude_modules", [])
        if not isinstance(exclude_modules_layer, list):
            exclude_modules_layer = []

        layers[layer_name] = LayerConfig(
            contains_modules=contains_modules,
            depends_on=depends_on_parsed,
            exclude_modules=exclude_modules_layer,
        )

    libs = {}
    for lib_name, lib_info in libs_config.items():
        # Check for allowed_in or upstream (backward compatibility)
        allowed_in = lib_info.get("allowed_in", lib_info.get("upstream", "none"))
        if allowed_in == "none":
            allowed_in_parsed = None
        elif isinstance(allowed_in, list):
            allowed_in_parsed = allowed_in
        else:
            allowed_in_parsed = None

        libs[lib_name] = LibConfig(allowed_in=allowed_in_parsed)

    return layers, libs, exclude_modules
